// -*- C++ -*- (c) 2007,2008 Petr Rockai <me@mornfall.net>

#include <QtCore/QObject>
#include <QtCore/QMap>
#include <QtCore/QSet>
#include <QtCore/QDebug>
#include <QtCore/QTimer>
#include <QtCore/QCoreApplication>

#include <QtGui/QAbstractItemDelegate>
#include <QtGui/QAbstractItemView>
#include <QtGui/QItemDelegate>
#include <QtGui/QLabel>
#include <QtGui/QTreeView>

#include <set>

#include <wibble/test.h>

#ifndef EPT_EXTENDABLELIST_H
#define EPT_EXTENDABLELIST_H

namespace adept {

class ExtendingEditorBuddy;

/**
 * @class ExtendingDelegate
 * Delegate that renders an extendable item.
 */
class ExtendingDelegate : public QItemDelegate {
    Q_OBJECT
private:
    /** Set of items that are currently extended and maps to the
     * editor widgets */
    typedef QSet< QPersistentModelIndex > IndexSet;
    mutable IndexSet m_extended;
    mutable QMap< QPersistentModelIndex, QWidget * > m_editors;

public:
    ExtendingDelegate( QObject *p = 0 ) : QItemDelegate( p ) {}

    IndexSet::iterator extendedBegin() { return m_extended.begin(); }
    IndexSet::iterator extendedEnd() { return m_extended.end(); }

    virtual QSize retractedSizeHint( const QStyleOptionViewItem &opt,
                                     const QModelIndex &idx ) const = 0;

    virtual QSize expandedSizeHint( const QStyleOptionViewItem &opt,
                                    const QModelIndex &idx ) const
    {
        QWidget *ed = m_editors[ idx ];
        int w = ed->width();
        int h = ed->heightForWidth( w );
        return QSize( w, retractedSizeHint( opt, idx ).height() + h );
    }

    /** Overrides QAbstractItem::sizeHint(opt,idx) to check if the
     * item is expanded. */
    virtual QSize sizeHint( const QStyleOptionViewItem &opt,
                            const QModelIndex &idx ) const {
        if ( m_extended.contains( idx ) )
            return expandedSizeHint( opt, idx );
        else
            return retractedSizeHint( opt, idx );
    }

    bool extended( const QModelIndex &i ) const {
        return m_extended.contains( i );
    }

    virtual QWidget *createEditorWidget( QWidget *parent,
                                         const QStyleOptionViewItem &opt,
                                         const QModelIndex &idx ) const
    {
        return 0;
    }

    /** Overrides QItemDelegate::createEditor
        Inserts the item into the set of extended items and creates an editor
        widget for it */
    virtual QWidget *createEditor( QWidget *parent,
                                   const QStyleOptionViewItem &opt,
                                   const QModelIndex &idx ) const {
        assert ( !m_extended.contains( idx ) );

        QWidget *w = createEditorWidget( parent, opt, idx );
        if ( !w )
            return 0;

        m_extended.insert( idx );
        QRect geom = opt.rect;
        geom.setTop( geom.top() + retractedSizeHint( opt, idx ).height() );
        w->setGeometry( geom );
        // w->setAutoFillBackground( true );
        m_editors[ idx ] = w;

        updateItem( idx );
        return w;
    }

    virtual void updateEditorGeometry (
        QWidget * editor,
        const QStyleOptionViewItem & opt,
        const QModelIndex & idx ) const
    {
        QRect old = editor->geometry();
        QRect geom = opt.rect;
        geom.setTop( geom.top() + retractedSizeHint( opt, idx ).height() );
        editor->setGeometry( geom );
        QSize hint = sizeHint( opt, idx );
        editor->resize( geom.size() );
        if ( old != editor->geometry() )
            editorGeometryUpdated();
    }

public Q_SLOTS:
    /** Removes the editor and its corresponding item's model index from the 
        extended items set */
    virtual void closeEditor( const QModelIndex &i ) {
        QObject *obj = m_editors[ i ];
        QWidget *w = qobject_cast< QWidget * >( obj );
        assert( w );
        m_extended.remove( i );
        m_editors.remove( i );
        updateItem( i );
    }
Q_SIGNALS:
    void editorGeometryUpdated() const;
    void updateItem( const QModelIndex & ) const;
};


class ExtendableListView : public QTreeView {
    Q_OBJECT
    friend class ExtendingEditorBuddy;
    
public:
    ExtendableListView( ExtendingDelegate *delegate, QWidget *parent = 0 );
    void closeAllEditors() {
        closeAllEditorsRequested();
    }

Q_SIGNALS:
    void closeAllEditorsRequested();
};


/**
 * @class ExtendingEditorBuddy
 * Helper class to help redrawing of extended delegates in the view.
 *
 * Create one and pass a view and delegate to the constructor.
 * It will tell the view to use the delegate, so
 * you don't call setItemDelegate on the view manually.
 * You don't need to delete the Buddy, since it
 * will be destroyed with the view automatically.
 */
class ExtendingEditorBuddy : public QObject {
    Q_OBJECT
private:
    ExtendableListView *m_view;
    ExtendingDelegate *m_delegate;
    /** List of items that need to be updated */
    QList< QPersistentModelIndex > m_update;
    bool m_updateDone;
    int m_updates;
    QTimer *m_updateTimer;

public:
    ExtendingEditorBuddy( ExtendableListView *v, ExtendingDelegate *d ) {
        m_updates = 0;
        m_updateTimer = new QTimer( this );
        m_view = v;
        connect( m_view, SIGNAL( closeAllEditorsRequested() ),
                 this, SLOT( closeAll() ) );
        m_delegate = d;
        v->setItemDelegate( d );
        assert( m_delegate );
        connect( v, SIGNAL( destroyed() ), this, SLOT( deleteLater() ) );

        connect( v, SIGNAL( clicked( const QModelIndex & ) ),
                 this, SLOT( toggleExtended( const QModelIndex & ) ) );

        connect( m_delegate, SIGNAL( editorGeometryUpdated() ),
                 this, SLOT( updateAll() ) );
        connect( m_delegate, SIGNAL( updateItem( const QModelIndex & ) ),
                 this, SLOT( update( const QModelIndex & ) ) );

        connect( m_updateTimer, SIGNAL( timeout() ),
                 this, SLOT( doUpdate() ) );
        m_updateTimer->setSingleShot( true );
    }

    ~ExtendingEditorBuddy() {
        delete m_delegate;
    }

public Q_SLOTS:
    virtual void update( const QModelIndex &m ) {
        m_update.append( m );
        if ( !m_updates )
            m_view->setUpdatesEnabled( false );
        ++ m_updates;
        m_updateTimer->start( 0 );
    }

protected Q_SLOTS:
    void doUpdate() {
        while ( !m_update.empty() ) { 
            -- m_updates;
            m_view->dataChanged( m_update.front(), m_update.front() );
            m_update.pop_front();
        }
        if ( !m_updates )
            m_view->setUpdatesEnabled( true );
        m_view->updateEditorGeometries();
        m_updateDone = true;
    }

    virtual void toggleExtended( const QModelIndex &m ) {
        if ( m_delegate->extended( m ) ) {
            m_view->closePersistentEditor( m );
            m_delegate->closeEditor( m );
        } else {
            m_view->openPersistentEditor( m );
            m_view->scrollTo( m );
        }
    }

    void closeAll() {
        std::set< QPersistentModelIndex >::iterator i;
        std::set< QPersistentModelIndex > toClose;

        std::copy( m_delegate->extendedBegin(), m_delegate->extendedEnd(),
                   std::inserter( toClose, toClose.begin() ) );

        for ( i = toClose.begin(); i != toClose.end(); ++i )
           toggleExtended( *i );

        if ( toClose.empty() )
            return;

        m_updateDone = false;
        while ( !m_updateDone )
            QCoreApplication::processEvents();
    }

    void updateAll() {
        QSet< QPersistentModelIndex >::iterator i;
        for ( i = m_delegate->extendedBegin();
              i != m_delegate->extendedEnd(); ++i )
            update( *i );
    }
};

}

#endif

