// -*- C++ -*- (c) 2008 Petr Rockai <me@mornfall.net>

#include <ept/core/apt.h>
#include <ept/core/xapian.h>
#include <ept/core/apt/action.h>

#include <adept/extendablelist.h>
#include <adept/util.h>
#include <adept/packageinfo.h>

#include <QtGui/QLabel>
#include <QtGui/QLayout>
#include <QtGui/QItemDelegate>
#include <QtGui/QPainter>
#include <QtGui/QPalette>
#include <QtGui/QCheckBox>

#include <klocale.h>
#include <kvbox.h>

#ifndef ADEPT_PACKAGELIST_H
#define ADEPT_PACKAGELIST_H

namespace adept {

using namespace ept::core;

class PackageExtender : public PackageInfo {
    Q_OBJECT

protected:
    QLabel *m_info, *m_desc;
    QWidget *m_box, *m_actBox;
    QVBoxLayout *m_layout;

public:
    PackageExtender( package::Source &_p, record::Source &_r,
                     package::ActionList &_actl, ept::Token _t,
                     QWidget *parent = 0 )
        : PackageInfo(_p, _r, _actl, _t, parent)
    {
        m_layout = new QVBoxLayout(this);
        QFontMetrics fm( font() );
        layout()->setContentsMargins( 12 + fm.height(), 5, 10, 10 );
        m_info = new QLabel( this );
        m_layout->addWidget(m_info);

        QLabel *filler = new QLabel( this ); // spacer
        m_layout->addWidget(filler);
        filler->setFixedSize( QSize( 1, 5 ) );

        m_desc = new QLabel( this );
        m_layout->addWidget(m_desc);
        m_info->setWordWrap( true );
        m_desc->setWordWrap( true );
        adjustFontSize( m_desc, -1 );

        filler = new QLabel( this ); // spacer
        m_layout->addWidget(filler);
        filler->setFixedSize( QSize( 1, 5 ) );

        m_box = new KHBox( this );
        m_layout->addWidget(m_box);
        m_actBox = new KVBox( m_box );

        refresh();
    }

public Q_SLOTS:
    virtual void refresh() {
        block = true;
        setUpdatesEnabled( false );

        QString info;
        PackageState s = p.get< package::State >( t );
        ept::Token candidate = p.get< package::CandidateVersion >( t ),
                   installed = p.get< package::InstalledVersion >( t );
        if ( s.upgradable() ) {
            info = i18n( "Version <b>" )
                   + u8( installed.version() )
                   + i18n( "</b> is installed, version <b>" )
                   + u8( candidate.version() )
                   + i18n( "</b> is available." );
        } else if ( !s.installed() && p.exists( candidate ) ) {
            info = i18n( "Version <b>" )
                   + u8( candidate.version() )
                   + i18n( "</b> is available." );
        } else if ( p.exists( installed ) ) {
            info = i18n( "Version <b>" )
                   + u8( installed.version() )
                   + i18n( "</b> is installed." );
        } else {
            info = i18n( "No versions available." );
        }

        info += i18n( " Package is maintained by <i>" )
                + u8( r.get< record::Maintainer >( t ) ) + i18n( "</i>" );
            
        m_desc->setText( formatLongDescription(
                             u8( r.get< record::LongDescription >( t ) ) ) );
        m_info->setText( info );

        if ( !m_act ) {
            m_act = new QCheckBox( m_actBox );
            connect( m_act, SIGNAL( stateChanged( int ) ), this, 
                     SLOT( actStateChanged( int ) ) );
        }

        if ( s.upgradable() ) {
            m_act->setText( i18n( "Upgrade this package" ) );
            if ( !m_remove ) {
                m_remove = new QCheckBox( m_actBox );
                m_remove->setText( i18n( "Remove this package" ) );
                connect( m_remove, SIGNAL( stateChanged( int ) ), this, 
                         SLOT( removeStateChanged( int ) ) );
            }
            m_act->setChecked( s.upgrade() );
            m_remove->setChecked( s.remove() );
        } else if ( s.installed() ) {
            delete m_remove;
            m_remove = 0;
            m_act->setText( i18n( "Remove this package" ) );
            m_act->setChecked( s.remove() );
        } else {
            delete m_remove;
            m_remove = 0;
            m_act->setText( i18n( "Install this package" ) );
            m_act->setChecked( s.install() );
        }
        setUpdatesEnabled( true );
        block = false;
        update();
    }

};

/**
 * @class PackageDelegate
 * Delegate that renders the description and status of a package
 * in an extendable list */
class PackageDelegate : public adept::ExtendingDelegate {
    Q_OBJECT

protected:
    record::Source &recs;
    package::Source &pkgs;
    package::ActionList &actl;

Q_SIGNALS:
    void changed();

public:
    ept::Token token( const QModelIndex &idx ) const {
        return idx.data( Qt::UserRole + 1 ).value< ept::Token >();
    }

    virtual QWidget *createEditorWidget( QWidget *parent,
                                         const QStyleOptionViewItem &opt,
                                         const QModelIndex &idx ) const
    {
        PackageExtender *e = new PackageExtender(
            pkgs, recs, actl, token( idx ), parent );
        connect( e, SIGNAL( changed() ), this, SIGNAL( changed() ) );
        connect( this, SIGNAL( changed() ), e, SLOT( refresh() ) );
        return e;
    }

    /** @param opt Parameters used to draw an item in a view widget.
     *  @param idx The model index of the item (not used). */
    virtual QSize retractedSizeHint( const QStyleOptionViewItem &opt,
                                     const QModelIndex &idx ) const
    {
        int h = QFontMetrics( opt.font ).height() + 2;
        return QSize( 200, 2 * h + 2 );
    }

    /** Paints the short/retracted package description */
    virtual void paint( QPainter *p, const QStyleOptionViewItem &_opt,
                        const QModelIndex &idx ) const
    {
        QStyleOptionViewItemV2 opt =
            static_cast< QStyleOptionViewItemV2 >( _opt );
        bool alternate = opt.features & QStyleOptionViewItemV2::Alternate;
        int width = opt.rect.width(),
           height = opt.rect.height();

        if ( width == 0 )
            return;

        QPalette pal = opt.palette;
        QColor ct = pal.color( QPalette::Current, QPalette::Text );
        
        QFont fb = opt.font;
        fb.setBold( true );
        QFontMetrics fm( opt.font );
        QFontMetrics fmb( fb );
        QBrush base = pal.brush( 
            QPalette::Current,
            alternate ? QPalette::AlternateBase : QPalette::Base );
        int size = fm.height() + 2; // icon size
        int off = size + 4;

        p->save();
        p->translate( opt.rect.left(), opt.rect.top() );
        p->setClipRect( 0, 0, width, height );
        p->setFont( opt.font );
        p->fillRect( 0, 0, width, height, QBrush( base ) );

        if ( opt.state & QStyle::State_Selected ) {
            p->fillRect( 0, 0, off, height, pal.color( QPalette::Current,
                                                       QPalette::Highlight ) );
        }

        off += 2;

        ept::Token t = token( idx );
        PackageState s = pkgs.get< package::State >( t );

        QString name = u8( t.package() ) + ": ",
              status = statusString( s ),
                dash = u8( " — " ),
              action = actionString( s ),
                desc = u8( recs.get< record::ShortDescription >( t ) );
        QSize szname = fmb.size( 0, name ),
            szstatus = fm.size( 0, status ),
              szdash = fm.size( 0, dash ),
            szaction = fm.size( 0, action ),
              szdesc = fm.size( 0, desc );
        szdesc.setWidth( std::min( szdesc.width(), width - off - 2 ) );
        desc = fm.elidedText( desc, Qt::ElideRight, szdesc.width() );
        
        int baseline = szname.height() - 2;
        
        p->setFont( fb );
        p->drawText( off, baseline, name );
        p->setFont( opt.font );
        p->drawText( off + 1, baseline + szdesc.height(), desc );
        off += szname.width();
        
        p->setPen( statusColor( s ) );
        p->drawText( off, baseline, status );
        off += szstatus.width();
        
        if ( !s.keep() ) {
            p->setPen( ct );
            p->drawText( off, baseline, dash );
            off += szdash.width();
            
            p->setPen( actionColor( s ) );
            p->drawText( off, baseline, action );
        }

        int y = ( szname.height() + szdesc.height() + 4 - size ) / 2;
        p->drawPixmap( 2, y, statusIcon( s, size ) );
        p->restore();
    }

    PackageDelegate( QObject *m,
                     package::Source &p,
                     record::Source &r,
                     package::ActionList &_actl) : pkgs( p ),
                                           recs( r ), actl( _actl )
    {
        connect( this, SIGNAL( changed() ), m, SLOT( refresh() ) );
    }
};

class PackageListView : public ExtendableListView {
    Q_OBJECT
    
public:
    PackageListView ( PackageDelegate *delegate, QWidget *parent = 0 )
        : ExtendableListView( delegate, parent ) {
        connect( this, SIGNAL( clicked( const QModelIndex & )),
            this, SLOT(clicked( const QModelIndex & )));
    }
    
    ept::Token token( const QModelIndex &idx ) const {
        return idx.data( Qt::UserRole + 1 ).value< ept::Token >();
    }
    
protected Q_SLOTS:
    void clicked( const QModelIndex &idx ) {
        emit selected(token(idx));
    }
    
Q_SIGNALS:
    void selected(ept::Token);
};


}

#endif
