// -*- C++ -*- (c) 2007, 2008 Petr Rockai <me@mornfall.net>

#include <adept/util.h>
#include <QtCore/QTimer>
#include <QtGui/QApplication>

#include <errno.h>

#include <apt-pkg/acquire.h>
#include <adept/pkgsystem.h>

// package display
#include <adept/packagelist.h>
#include <adept/tokenmodel.h>
#include <adept/packagedetails.h>

// download/commit-time widgets
#include <adept/downloadprogress.h>
#include <adept/guidpkgpm.h>
#include <adept/commitwidget.h>

// general-purpose widgets
#include <adept/sidebar.h>
#include <QtGui/QPushButton>
#include <QtGui/QHeaderView>
#include <QtGui/QStackedWidget>
#include <QtGui/QProgressBar>
#include <KXmlGuiWindow>
#include <KIcon>
#include <KVBox>
#include <KHBox>
#include <KLineEdit>

#ifndef ADEPT_MANAGER_H
#define ADEPT_MANAGER_H

using namespace ept::core;
using namespace adept;

class OpProgressWidget : public KVBox, public OpProgress
{
    Q_OBJECT
public:
    OpProgressWidget( QWidget *p )
        : KVBox( p ), m_busy( false )
    {
        QSizePolicy pol( QSizePolicy::Preferred,
                         QSizePolicy::Fixed );
        new QWidget( this ); // spacer
        m_label = new QLabel( this );
        m_label->setSizePolicy( pol );
        m_pbar = new QProgressBar( this );
        m_pbar->setMaximum( 100 );
        m_pbar->setSizePolicy( pol );
        new QWidget( this ); // spacer
    }

    virtual void Update() {
        if ( CheckChange( 0.05 ) == false )
            return;
        if (MajorChange) {
            if ( !m_busy ) {
                emit activated();
                QApplication::setOverrideCursor( QCursor( Qt::BusyCursor ) );
                m_busy = true;
            }
        }
        m_label->setText( l8bit( Op ) + "..." );
        m_pbar->setValue( Percent );
        QCoreApplication::processEvents();
    }

    virtual void Done() {
        QApplication::restoreOverrideCursor();
        m_busy = false;
        emit done();
    }

Q_SIGNALS:
    void activated();
    void done();
protected:
    QLabel *m_label;
    QProgressBar *m_pbar;
    bool m_busy;
};

class ProgressWidget : public QStackedWidget
{
    Q_OBJECT
    CommitWidget *m_commit;
    DownloadProgress *m_download;
    OpProgressWidget *m_progress;
    QLabel *m_noop;

public:
    ProgressWidget( QWidget *p = 0 ) : QStackedWidget( p )
    {
        m_noop = new QLabel( this );
        addWidget( m_noop );
        m_noop->setWordWrap( true );
        m_noop->setMargin( 10 );

        m_progress = new OpProgressWidget( this );
        addWidget( m_progress );

        m_commit = new CommitWidget( this );
        addWidget( m_commit );

        m_download = new DownloadProgress( this );
        addWidget( m_download );

        connect( m_download,
                 SIGNAL( statusChanged( DownloadProgress::Status ) ),
                 this,
                 SLOT( downloadStatusChanged( DownloadProgress::Status ) ) );
        connect( m_commit, SIGNAL( activated() ),
                 this, SLOT( raiseCommit() ) );

        connect( m_progress, SIGNAL( activated() ),
                 this, SLOT( progressStart() ) );
        connect( m_progress, SIGNAL( done() ),
                 this, SLOT( progressDone() ) );

        noop();
    }

    DebconfGui *debconf() {
        return m_commit->debconf();
    }

    DownloadProgress *download() {
        return m_download;
    }

    OpProgressWidget *progress() {
        return m_progress;
    }

    CommitWidget *commit() {
        return m_commit;
    }

public Q_SLOTS:
    void downloadStatusChanged( DownloadProgress::Status s ) {
        if ( s == DownloadProgress::StWaiting )
            setCurrentWidget( m_download );
    }

    void raiseCommit() {
        setCurrentWidget( m_commit );
    }

    void progressStart() {
        setCurrentWidget( m_progress );
    }

    void progressDone() {
        setCurrentWidget( m_noop );
    }

    void noop() {
        setCurrentWidget( m_noop );
        m_noop->setAlignment( Qt::AlignCenter );
        m_noop->setText( i18n( "No operation currently in progress. "
                             "Please go to Preview to apply any pending "
                             "changes or to Sources to update available "
                             "package lists." ) );
    }

    void done() {
        setCurrentWidget( m_noop );
        m_noop->setAlignment( Qt::AlignCenter | Qt::AlignVCenter );
        m_noop->setText(
            i18n( "The operation finished successfully and "
                  "no operation is currently in progress. "
                  "You may now exit the program or go back to "
                  "installing and removing packages." ) );
    }

    void downloadCancelled() {
        setCurrentWidget( m_noop );
        m_noop->setAlignment( Qt::AlignCenter | Qt::AlignVCenter );
        m_noop->setText(
            i18n( "You have cancelled package download, therefore the"
                  " installation or upgrade that was requested cannot"
                  " be completed. I have tried to preserve the operation"
                  " you have requested, so you could return to the Preview"
                  " tab and click Apply changes again to restart the"
                  " operation. Otherwise, you may exit the program or go"
                  " back to installing and removing packages." ) );
    }

    void error() {
        setCurrentWidget( m_noop );

        QString errs;
        while ( !_error->empty() ) {
            std::string err;
            _error->PopMessage( err );
            std::cerr << err << std::endl;
            errs += u8( err ) + "<br>";
        }

        m_noop->setAlignment( Qt::AlignJustify | Qt::AlignVCenter );
        m_noop->setText(
            i18n( "The operation requested has failed. Unfortunately, no"
                  " error recovery is currently implemented, please investigate"
                  " the situation manually. It may be useful to issue"
                  " <b>dpkg --configure -a</b> (as root, or through sudo)."
                  " <br><b>%1</b>", errs ) );
    }
};

class SourcesWidget : public KVBox
{
    Q_OBJECT
protected:
    QLabel *m_label;
    QPushButton *m_update;
public:
    SourcesWidget( QWidget *p = 0 )
        : KVBox( p )
    {
        setMargin( 10 );
        layout()->setSpacing( 3 );

        new QWidget(); // spacer

        // TODO if software-properties-kde is installed, offer to run it?
        // TODO implement apt-get update at least, it's not that hard
        m_label = new QLabel( this );
        m_label->setText(
            i18n( "The editor for package sources is <b>not yet implemented</b>."
                  " Please use the <b>Software Properties</b> tool in the"
                  " meantime. (You need to close Adept first and re-start it"
                  " after setting up your sources, though. To refresh the list"
                  " of available software, click the button below." ) );
        m_label->setWordWrap( true );
        m_label->setAlignment( Qt::AlignCenter );

        m_update = new QPushButton( this );
        m_update->setText( i18n( "Fetch current package lists" ) );
        connect( m_update, SIGNAL( clicked() ),
                 this, SIGNAL( fetchUpdates() ) );

        new QWidget(); // spacer
    }

Q_SIGNALS:
    void fetchUpdates();
};

class AdeptManager : public KXmlGuiWindow
{
    Q_OBJECT

    AptDatabase db;
    package::Source pkgs;
    record::Source recs;
    xapian::Source xap;
    package::ActionList actl;

    TokenModel *m_searchModel, *m_previewModel;
    QStackedWidget *m_stack;

    KVBox *m_transient;
    QLabel *m_transientLabel;
    QPushButton *m_transientOk;

    Sidebar *m_sidebar;
    KVBox *m_search, *m_preview;
    KHBox *m_previewButtons;
    KLineEdit *m_searchLine;
    PackageListView *m_searchList, *m_previewList;
    QPushButton *m_applyButton, *m_revertButton, *m_upgradeButton;
    
    PackageDetails *m_details;
    ProgressWidget *m_progress;
    QTimer *m_searchTimer;
    bool m_longTimeout;

    int m_interruptedWidget;

public:

    PackageListView *setupList( QWidget *p, TokenModel *m )
    {
        PackageListView *r = new PackageListView(
            new PackageDelegate( this, pkgs, recs, actl ), p );
        r->setModel( m );
        r->header()->hide();
        return r;
    }

    void transient( QString txt ) {
        m_stack->setCurrentWidget( m_transient );
        m_transientLabel->setText( txt );
    }

    AdeptManager( QWidget *p = 0 ) : KXmlGuiWindow( p ),
                                     pkgs( db ),
                                     recs( db )
    {
        new PkgSystem( db );

        m_stack = new QStackedWidget( this );

        m_transient = new KVBox( m_stack );
        m_transient->setMargin( 10 );
        m_stack->addWidget( m_transient );

        new QWidget( m_transient );
        m_transientLabel = new QLabel( m_transient );
        m_transientLabel->setAlignment( Qt::AlignJustify | Qt::AlignVCenter );
        m_transientLabel->setWordWrap( true );
        m_transientOk = new QPushButton( m_transient );
        m_transientOk->setText( i18n( "Ok" ) );
        connect( m_transientOk, SIGNAL( clicked() ),
                 this, SLOT( transientClose() ) );
        new QWidget( m_transient );

        m_sidebar = new Sidebar( m_stack );
        m_stack->addWidget( m_sidebar );
        m_stack->setCurrentWidget( m_sidebar );

        m_search = new KVBox();
        m_preview = new KVBox();
        m_details = new PackageDetails(pkgs,recs,actl);

        m_searchLine = new KLineEdit( m_search );
        m_searchLine->setClearButtonShown( true );
        m_searchModel = new TokenModel();
        m_searchList = setupList( m_search, m_searchModel );

        m_previewModel = new TokenModel();
        m_previewList = setupList( m_preview, m_previewModel );
        m_previewButtons = new KHBox( m_preview );

        m_upgradeButton = new QPushButton( m_previewButtons );
        m_revertButton = new QPushButton( m_previewButtons );
        m_applyButton = new QPushButton( m_previewButtons );
        m_upgradeButton->setText( i18n( "Upgrade" ) );
        m_revertButton->setText( i18n( "Revert changes" ) );
        m_applyButton->setText( i18n( "Apply changes" ) );

        m_progress = new ProgressWidget();
        db.setProgress( m_progress->progress() );

        SourcesWidget *sources = new SourcesWidget();

        m_sidebar->addItem( sources, KIcon( "applications-internet" ),
                            "Sources" );
        m_sidebar->addItem( m_search, KIcon( "edit-find" ), "Search" );
        m_sidebar->addItem( m_details, KIcon( "edit-copy" ), "Details" );
        m_sidebar->addItem( m_preview, KIcon( "dialog-ok-apply" ), "Changes" );
        m_sidebar->addItem( m_progress, KIcon( "system-run" ), "Progress" );

        setCentralWidget( m_stack );
        m_sidebar->setCurrentIndex( 1 );
        m_search->setFocusProxy( m_searchLine );
        m_search->setFocus( Qt::OtherFocusReason );

        QTimer::singleShot( 0, this, SLOT( fill() ) );

        connect( m_searchLine, SIGNAL( returnPressed() ), this, SLOT( fill() ) );
        connect( m_searchList, SIGNAL( selected(ept::Token) ),
            m_details, SLOT(setPackage(ept::Token)));

        connect( m_applyButton, SIGNAL( clicked() ), this, SLOT( apply() ) );
        connect( m_revertButton, SIGNAL( clicked() ), this, SLOT( revert() ) );
        connect( m_upgradeButton, SIGNAL( clicked() ), this, SLOT( upgrade() ) );

        // set up search-as-you-type
        m_longTimeout = false;
        m_searchTimer = new QTimer();
        m_searchTimer->setSingleShot( true );
        connect( m_searchLine, SIGNAL( textChanged( const QString & ) ),
                 this, SLOT( searchTextChanged() ) );
        connect( m_searchTimer, SIGNAL( timeout() ),
                 this, SLOT( searchTextTimeout() ) );

        connect( m_progress->progress(), SIGNAL( activated() ),
                 this, SLOT( progressStart() ) );
        connect( m_progress->progress(), SIGNAL( done() ),
                 this, SLOT( progressDone() ) );

        connect( sources, SIGNAL( fetchUpdates() ),
                 this, SLOT( update() ) );

        checkAptDatabase();
    }

public slots:
    void progressStart() {
        m_interruptedWidget = m_sidebar->currentIndex();
        m_sidebar->setCurrentIndex( 4 );
    }

    void progressDone() {
        m_sidebar->setCurrentIndex( m_interruptedWidget );
    }

    void checkAptDatabase() {
        disconnect( this, SIGNAL( transientClosed() ),
                    this, SLOT( checkAptDatabase() ) );
        db.invalidate();
        try {
            db.cache(); // this here forces cacheOpen()
        } catch ( RecoveryNeededException &e ) {
            connect( this, SIGNAL( transientClosed() ),
                     this, SLOT( runRecovery() ) );
            transient(
                i18n( "It appears that dpkg has been interrupted and there"
                      " are some operations that need to be finished before"
                      " normal operation can be resumed. If you say OK here,"
                      " I can try to replay the failed operations. If this"
                      " fails, it is advisable that you investigate the"
                      " situation manually, using <b>dpkg --configure -a</b>"
                      " as root, in a terminal. If you prefer that Adept"
                      " does not attempt recovery, you can safely close the"
                      " program now." ) );
        } catch ( LockingFailureException &e ) {
            transient(
                i18n( "Could not obtain a write lock on the cache, "
                      " falling back to <b>read-only mode</b>. "
                      " You won't be able to install, remove or upgrade"
                      " packages. However, you can still search in"
                      " the package database and browse packages.<br> %1",
                      e.code() == EACCES || e.code() == EPERM ?
                      i18n(
                          "You apparently do not have sufficient permissions to"
                          " install or remove programs. You may need to run this"
                          " program as user <b>root</b> (or through <b>sudo</b>)"
                          " to gain write access. " ) :
                      i18n( 
                          "It appears that another process is running, which"
                          " holds the write lock on the database. You first need"
                          " to close that program and then restart Adept to gain"
                          " write access." ) ) );
        } catch ( std::exception &e ) {
            transient(
                i18n( "Something broke. Please close the application and"
                      " try to do something about it... The exception says: %1",
                      e.what() ) );
        }
        m_progress->progress()->Done();
    }

    void runRecovery() {
        disconnect( this, SIGNAL( transientClosed() ),
                    this, SLOT( runRecovery() ) );
        DpkgGui *dpkg = new DpkgGui( pkgs );
        dpkg->setDebconf( m_progress->debconf() );
        m_progress->commit()->setPM( dpkg );
        m_progress->commit()->setStatus( 0, i18n( "Starting recovery..." ) );
        m_progress->raiseCommit();
        m_sidebar->setCurrentIndex( 4 );
        if ( dpkg->recover() ) {
            m_sidebar->setCurrentIndex( 1 );
            connect( this, SIGNAL( transientClosed() ),
                     this, SLOT( checkAptDatabase() ) );
            transient(
                i18n( "It seems that the recovery has been successfuly. We will"
                      " now try to re-open the cache read-write..." ) );
        }
    }

    void transientClose() {
        m_stack->setCurrentWidget( m_sidebar );
        emit transientClosed();
    }

    void searchTextChanged() {
        m_longTimeout = false;
        m_searchTimer->start( 200 );
    }

    void searchTextTimeout() {
        if ( m_longTimeout ) {
            fill();
        } else {
            m_longTimeout = true;
            m_searchTimer->start( 300 );
            partialFill();
        }
    }

    void refresh() {
        m_searchList->viewport()->repaint();
        m_previewList->viewport()->repaint();
        fillPreview();
    }

    void actionListChanged() {
        pkgs.revertStates();
        actl.replay( pkgs );
        fillPreview();
    }

    void revert() {
        actl.clear();
        actionListChanged();
    }

    void upgrade() {
        actl.add( package::Action(
                      ept::Token(), package::Action::SystemUpgrade ) );
        actionListChanged();
    }

    void update() {
        pkgSourceList SList;

        if(!SList.ReadMainList())
            abort();

        m_sidebar->setCurrentIndex( 4 );

        FileFd Lock;
        Lock.Fd(GetLock(_config->FindDir("Dir::State::Lists") + "lock"));
        // TODO check for failure

        std::auto_ptr< pkgAcquire > fetcher(
            new pkgAcquire( m_progress->download() ) );
        SList.GetIndexes( fetcher.get() );

        pkgAcquire::RunResult res =
#ifndef RPM
            fetcher->Run( 50000 );
#else
            fetcher->Run();
#endif

       if( res == pkgAcquire::Failed )
           abort();

       if( res == pkgAcquire::Cancelled )
           return; // do nothing?

        bool Failed = false;
        string FailedURI;
        for (pkgAcquire::ItemIterator I = fetcher->ItemsBegin();
             I != fetcher->ItemsEnd(); I++)
        {
            if ((*I)->Status == pkgAcquire::Item::StatDone)
                continue;

            (*I)->Finished();

            if((*I)->ErrorText.empty())
                FailedURI += (*I)->DescURI() + "\n";
            else
                FailedURI += (*I)->DescURI() + ": " + (*I)->ErrorText + "\n";
            Failed = true;
        }

        // Clean out any old list files
        if (_config->FindB("APT::Get::List-Cleanup",true) == true)
        {
            if ( fetcher->Clean(_config->FindDir("Dir::State::lists")) == false ||
                 fetcher->Clean(_config->FindDir("Dir::State::lists")
                                  + "partial/") == false )
                ; //return false;
        }

        m_progress->done();
        db.invalidate();
    }

    void apply() {
        bool invalidated = false;
        pkgPackageManager *_pm = _system->CreatePM( &db.state() );
        std::auto_ptr< pkgAcquire > fetcher(
            new pkgAcquire( m_progress->download() ) );
        pkgSourceList SList;

        GuiDPkgPM *pm = dynamic_cast< GuiDPkgPM * >( _pm );
        assert( pm );
        pm->harness().setDebconf( m_progress->debconf() );

        m_progress->commit()->setPM( &pm->harness() );
        m_progress->commit()->setStatus( 0, i18n( "Initializing..." ) );
        m_sidebar->setCurrentIndex( 4 );

        if(!SList.ReadMainList())
            abort();

        std::auto_ptr< pkgRecords > r( new pkgRecords( db.cache() ) );

        while (1) {

            // TODO we need to check _error here, possibly (as well as
            // in some other places?)
            fetcher->Shutdown(); // reload
            if ( !pm->GetArchives( fetcher.get(), &SList, r.get() ) )
                abort();

            pkgAcquire::RunResult res = 
#ifndef RPM
                fetcher->Run( 50000 );
#else
                fetcher->Run();
#endif
                
            if ( res == pkgAcquire::Failed ) {
                _error->Error( "Package download failed." );
                m_progress->error();
                break;
            }

            if ( res == pkgAcquire::Cancelled ) {
                m_progress->downloadCancelled();
                break;
            }

            /* TODO check that we managed to download everything,
             * otherwise we wind up with Internal error */

            _system->UnLock();
#ifndef RPM
            pkgPackageManager::OrderResult Res = pm->DoInstall(-1);
#else
            pkgPackageManager::OrderResult Res = pm->DoInstall();
#endif
            _system->Lock();
            invalidated = true;

            if (Res == pkgPackageManager::Completed) {
                m_progress->done();
                break;
            } else if (Res == pkgPackageManager::Failed) {
                m_progress->error();
                break;
            }

        }
        delete pm;

        if ( invalidated ) {
            db.invalidate();
            actl.prune( pkgs );
            actionListChanged();
        }
    }

    // TODO in both fill methods, we need to prevent db.invalidate()
    // from happening (we will get crashes otherwise)
    // ... OK, probably not crashes in the fill() case, but still...

    void fillPreview() {
        m_previewList->closeAllEditors();
        m_previewModel->fillFrom( pkgs.changedList() );
    }

    void fill( bool partial = false ) {
        QString txt = m_searchLine->text();
        if ( txt.length() < 3 ) {
            m_searchModel->clear();
            return;
        }
        // maybe use toUtf8 intsead?
        std::string query = txt.toLocal8Bit().data();
        m_searchList->closeAllEditors();
        m_searchModel->fillFrom(
            partial ? xap.partialQuery( query ) : xap.query( query ) );
    }

    void partialFill() {
        fill( true );
    }
Q_SIGNALS:
    void transientClosed();
};

#endif
